{ splitter.pas
    Implements a splitter window. Windows >3.1x only!

    Author: Alfons Hoogervorst.
    No copyrights claimed. Donated to the public domain.
    (Hey! I really won't mind my name appearing in credits.
    It's such a nice name :-)

    E-mail:
    Fido 2:500/121.6252
    Internet a.hoogervorst@dosgg.nl

    Last modifications: July-August 1995
}

unit Splitter;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus;

type
  { This class to work around a bug(?). Modified property doesn't
    change after changing the bitmap contents }
  TBrushBitmap = class(TBitmap)
  protected
    IsChanged: Boolean;
    property OnChange;
    procedure Handle_OnChange(Sender: TObject);
  end;

  TOrientation = (orHorizontal, orVertical);

  { During window creation the splitter window fires a OnSplitBrush event.
    Use this event to supply your own splitwindow drag bar. The first
    param (Sender) identifies the object that fired the event. The
    second param is a "brush canvas". Areas that are painted black in
    the canvas will be inverted in the drag bar. White areas will
    be transparent. SplitBarWidth is just the width of the SplitBar when
    dragged. Change this variable to create drag bars with screen
    independent dimensions (by using GetSystemMetrics or metrics of
    a system font). }
  TSplitBrushEvent = procedure (Sender: TObject; BrushCanvas: TCanvas;
    var SplitBarWidth: Integer) of object;

  TSplitterChangeEvent = procedure (Sender: TObject; OffsX, OffsY: Integer) of object;
  TSplitter = class(TCustomControl)
  private
    { Private declarations }
    FBorderStyle: TBorderStyle;     { Border style }
    FOrientation: TOrientation;     { Splitter type }
    FSplitterWidth: Integer;
    FOnChange: TSplitterChangeEvent;
    FOnPaint: TNotifyEvent;
    FOnSplitBrush: TSplitBrushEvent;
    FSplitting: Word;                { Flags splitting's busy }
    FClipRect: TRect;

    procedure CalcLastPos(var pos: TPoint);
    procedure RestrictPos(var pos: TPoint);
    procedure SetBorderStyle(style: TBorderStyle);
    procedure SetOrientation(SplitterOrientation: TOrientation);
    function GetSplitting: Boolean;
    procedure SetSplitting(IsSplitting: Boolean);
    function GetSplittingByMouse: Boolean;
    procedure SetSplittingByMouse(SplittingByMouse: Boolean);

  protected
    { Protected declarations }
    OldPos, LastPos, OffsetPos: TPoint;      { Internal storage of last position }

    SplitBrush: TBrush;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Paint; override;
    procedure CreateWnd; override;
    procedure DrawSplitter(pt: TPoint); virtual;
    procedure Fire_OnChange; dynamic;
    procedure Fire_OnSplitBrush; dynamic;

    { Message methods }
    procedure WMLButtonDown(var Message: TMessage); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TMessage); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TMessage); message WM_MOUSEMOVE;
    procedure WMGetDlgCode(var Message: TMessage); message WM_GETDLGCODE;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure DraggingStarts(ByMouse: Boolean); dynamic;
    procedure DraggingToPoint(pt: TPoint); dynamic;
    procedure DraggingToOffset(x, y: Integer); dynamic;
    procedure DraggingEnds(CancelDrag: Boolean); dynamic;
    procedure Recreate;

    function GetControlRect: TRect;
    function GetControlOrigin: TPoint;
    function GetParentClientRect: TRect;
    function GetParentClientOrigin: TPoint;

    property Canvas;             { For OnPaint }
    property MouseCapture;       { For capturing mouse movement }
    property SplitterWidth: Integer read FSplitterWidth write FSplitterWidth;
    property Splitting: Boolean read GetSplitting write SetSplitting;
    property SplittingByMouse: Boolean read GetSplittingByMouse write
      SetSplittingByMouse;
    property ClipRect: TRect read FClipRect write FClipRect;

  published
    { Published declarations }
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle
      default bsSingle;
    property Color;
    property OnChange: TSplitterChangeEvent read FOnChange write FOnChange;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnSplitBrush: TSplitBrushEvent read FOnSplitBrush write FOnSplitBrush;
    property Orientation: TOrientation read FOrientation write SetOrientation
      default orVertical;
    property ParentColor;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
  end;

procedure Register;

implementation

{ TBrushBitmap.Handle_OnChange - Handles chaning of bitmap }
procedure TBrushBitmap.Handle_OnChange(Sender: TObject);
begin
  IsChanged := True;
end;

{ TSplitter.Create - Creates SplitBrush }
constructor TSplitter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FBorderStyle := bsSingle;
  FOrientation := orVertical;
  if FOrientation = orHorizontal then Cursor := crVSplit
  else Cursor := crHSplit;
  SplitBrush := TBrush.Create;
  { All other are zeroed }
end;

destructor TSplitter.Destroy;
begin
  SplitBrush.Free;
  inherited Destroy
end;

{ TSplitter.CreateParams - Before splitter window (the handle!) is created
    we fiddle a little with style bits to make the border style OK. }
procedure TSplitter.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if FBorderStyle = bsSingle then Params.Style := Params.Style or WS_BORDER;
end;

{ TSplitter.Paint - Fires an OnPaint event if available }
procedure TSplitter.Paint;
begin
  if Assigned(FOnPaint) then FOnPaint(Self)
  else inherited Paint
end;

{ TSplitter.CreateWnd - Necessary for modifying properties/fields
    that depend on the window associated with TSplitter. Also fires the
    OnSplitBrush event. }
procedure TSplitter.CreateWnd;
begin
  inherited CreateWnd;
  if FOrientation = orHorizontal then FSplitterWidth := Height
  else FSplitterWidth := Width;

  { Before we continue we send an OnSplitBrush event }
  Fire_OnSplitBrush
end;

{ TSplitter.Fire_OnChange - I've written quite a lot VBX controls.
    "Firing" events sounds appropriate for custom controls. }
procedure TSplitter.Fire_OnChange;
begin
  if Assigned(FOnChange) then
    FOnChange(Self, LastPos.x - OldPos.x, LastPos.y - OldPos.y);
end;

{ TSplitter.Fire_OnSplitBrush - Creates the brush for the dragged
    split bar. }
procedure TSplitter.Fire_OnSplitBrush;
begin
  { Recreate brush }
  SplitBrush.Free;
  SplitBrush := TBrush.Create;

  { Recreate bitmap in brush }
  SplitBrush.Bitmap.Free; { Destroy old one if available }
  SplitBrush.Bitmap := TBrushBitmap.Create;
  SplitBrush.Bitmap.Width := 8; SplitBrush.Bitmap.Height := 8;

  if Assigned(FOnSplitBrush) then
  begin
    { Install our OnChange event handler }
    (SplitBrush.Bitmap as TBrushBitmap).IsChanged := False;
    SplitBrush.Bitmap.OnChange := (SplitBrush.Bitmap as TBrushBitmap).Handle_OnChange;
    FOnSplitBrush(Self, SplitBrush.Bitmap.Canvas, FSplitterWidth);
    SplitBrush.Bitmap.OnChange := nil;
  end;
  { Set to default if bitmap did NOT change }
  if not (SplitBrush.Bitmap as TBrushBitmap).IsChanged then
  begin
    SplitBrush.Bitmap.Canvas.Brush.Color := clBlack;
    SplitBrush.Bitmap.Canvas.FillRect(Rect(0, 0,
      SplitBrush.Bitmap.Width, SplitBrush.Bitmap.Height));
  end;
  { SplitBrush is protected. Derived classes might mess with this brush
    so IsChanged should be set to False }
  (SplitBrush.Bitmap as TBrushBitmap).IsChanged := False
end;

{ TSplitter.WMLButtonDown - If we're receiving a WM_LBUTTONDOWN it's
    time to move the splitter line around }
procedure TSplitter.WMLButtonDown(var Message: TMessage);
begin
  DraggingStarts(True);
  inherited
end;

{ TSplitter.WMLButtonUp - OK, button's up. Undo all splitting and make
    splitter window move. Fire OnChange event. }
procedure TSplitter.WMLButtonUp(var Message: TMessage);
begin
  DraggingEnds(False);
  inherited
end;

{ TSplitter.WMMouseMove - Mouse moves while dragging split bar? }
procedure TSplitter.WMMouseMove(var Message: TMessage);
var
  MousePos: TPoint; { WM_MOUSEMOVE }
begin
  if SplittingByMouse then
  begin
    GetCursorPos(MousePos);
    DraggingToPoint(MousePos);
  end;
  inherited
end;

{ TSplitter.OnGetDlgCode - This TSplitter window works with mouse only.
    Change dlg code to suit your needs. Seasoned custom control writers
    shouldn't have much trouble with this stuff. }
procedure TSplitter.WMGetDlgCode(var Message: TMessage);
begin
  Message.Result := DLGC_STATIC
end;

{ TSplitter.DrawSplitter - Draws the dragged split bar by using the
    SplitBrush. Uses the DPxn ROP (Raster OPeration code) to invert
    the black areas in the brush. White areas in brush will be
    transparent. How did I determine the ROP code? By using a trick.
    Cf. one of my messages in WINDOWS.PROG. }
procedure TSplitter.DrawSplitter(pt: TPoint);
const
  { Xor Destination  with Pattern, invert result }
  DPxn = $00A50065;
var
  hdcScreen: HDC;
  Where: TPoint;
  rcRect: TRect;
begin
  rcRect := GetControlRect;
  if FOrientation = orVertical then
  begin
    Where.x := pt.x; Where.y := rcRect.top;
    rcRect.right := FSplitterWidth; dec(rcRect.bottom, rcRect.top);
  end else
  begin
    Where.x := rcRect.left; Where.y := pt.y;
    Dec(rcRect.right, rcRect.left); rcRect.bottom := FSplitterWidth;
  end;
  { Okay, draw split brush. I do it on the screen DC. Take a look at
    the GetWindowDC API. If you're using the Window DC, don't forget
    to change the GetControlxxx members. Work with screen coordinates. }
  hdcScreen := GetDC(0);
  UnrealizeObject(SplitBrush.Handle);
  SetBrushOrg(hdcScreen, Where.x, Where.y);
  SelectObject(hdcScreen, SplitBrush.Handle);
  PatBlt(hdcScreen, Where.x, Where.y, rcRect.right, rcRect.bottom, DPxn);
  ReleaseDC(0, hdcScreen);
end;

{ TSPlitter.DraggingStarts - Initiates dragging of split bar.
    Pass true if a mouse action initiated the drag. I created the
    Draggingxxx functions so you can initiate  dragging from
    within other handlers/classes/components (presumably for
    groups of TSplitters or keyboard driven ;-) splitters  }
procedure TSplitter.DraggingStarts(ByMouse: Boolean);
begin
  if not Splitting then
  begin
    Splitting := True;
    { Set by mouse flag & calculate offset  }
    if ByMouse then
    begin
      SplittingByMouse := True;
      GetCursorPos(LastPos);
    end else LastPos := GetControlOrigin;
    OffsetPos := GetControlOrigin;
    OldPos := OffsetPos;
    MapWindowPoints(HWND_DESKTOP, Parent.Handle, OldPos, 1);
    OffsetPos := Point(LastPos.x - OffsetPos.x, LastPos.y - OffsetPos.y);

    { Now recalc, restrict and adjust LastPos }
    if ByMouse then GetCursorPos(LastPos);
    CalcLastPos(LastPos);
    if ByMouse then MouseCapture := True;
    DrawSplitter(LastPos); { Draw line at current position }
  end;
end;

{ TSplitter.DraggingToPoint - Drags split bar to a point (in screen
   coordinates). If you're dragging with mouse pass the current cursor
   position. Else use DraggingToOffset (preferred) or do the calcs
   yourself (not preferred, a major hassle). }
procedure TSplitter.DraggingToPoint(pt: TPoint);
begin
  if Splitting then
  begin
    CalcLastPos(pt);
    if ((FOrientation = orHorizontal) and (pt.y <> LastPos.y)) or
       ((FOrientation = orVertical)   and (pt.x <> LastPos.x)) then
    begin
      DrawSplitter(LastPos);  { Clear old line at LastPos }
      DrawSplitter(pt); { Draw new line }
      LastPos := pt
    end
  end
end;

{ TSplitter.DraggingToOffset - Moves the dragged bar relative to the
    previous position. }
procedure TSplitter.DraggingToOffset(x, y: Integer);
begin
  DraggingToPoint(Point(LastPos.x + x, LastPos.y + y));
end;

{ TSplitter.DraggingEnds - Terminates dragging process. Remember:
    DraggingEnds fires the OnChange event! So call this function
    if you want to terminate the dragging and fire a OnChange
    event! If CancelDrag = True, the splitter cancels the dragging
    process. }
procedure TSplitter.DraggingEnds(CancelDrag: Boolean);
begin
  if Splitting then
  begin
    if SplittingByMouse then MouseCapture := False;
    DrawSplitter(LastPos); { Clear out old line }

    if not CancelDrag then
    begin
      { Calculate new position }
      if SplittingByMouse then GetCursorPos(LastPos);
      CalcLastPos(LastPos);
      MapWindowPoints(HWND_DESKTOP, Parent.Handle, LastPos, 1);

      { Move }
      if FOrientation = orHorizontal then Top := LastPos.y
      else Left := LastPos.x;

      { Fire event }
      Fire_OnChange;
    end;

    Splitting := False;
  end
end;

{ TSplitter.Recreate - Recreates TSpitter window. Might be handy after
    Modifying the splitter brush }
procedure TSplitter.Recreate;
begin
  RecreateWnd;
end;

procedure TSplitter.CalcLastPos(var pos: TPoint);
begin
  pos := Point(pos.x - OffsetPos.x, pos.y - OffsetPos.y);
  RestrictPos(pos)
end;

{ TSplitter.RestrictPos - Ensures that last pos is always within the clip
    rectangle. }
procedure TSplitter.RestrictPos(var pos: TPoint);
var
  rcClipScreen: TRect;
begin
  if IsRectEmpty(FClipRect) then FClipRect := Parent.ClientRect;
  rcClipScreen := FClipRect;
  MapWindowPoints(Parent.Handle, HWND_DESKTOP, rcClipScreen, 2);
  case FOrientation of
    orHorizontal:
      if pos.y < rcClipScreen.top then pos.y := rcClipScreen.top
      else if (pos.y + FSplitterWidth) > rcClipScreen.bottom then
        pos.y := rcClipScreen.bottom - FSplitterWidth;
    orVertical:
      if pos.x < rcClipScreen.left then pos.x := rcClipScreen.left
      else if (pos.x + FSplitterWidth) > rcClipScreen.right then
        pos.x := rcClipScreen.right - FSplitterWidth;
  end
end;

procedure TSplitter.SetBorderStyle(style: TBorderStyle);
var
  fRedraw: Boolean;
begin
  fRedraw := style <> FBorderStyle;
  if fRedraw then
  begin
    FBorderStyle := style;
    if Handle <> 0 then RecreateWnd
  end;
end;

{ TSplitter.SetOrientation - Changes orientation of splitter. }
procedure TSplitter.SetOrientation(SplitterOrientation: TOrientation);
begin
  if FOrientation <> SplitterOrientation then
  begin
    FOrientation := SplitterOrientation;
    if FOrientation = orHorizontal then Cursor := crVSplit
    else Cursor := crHSplit;

    { We're calling RecreateWnd so the control will send a new
      SplitterBrush event. }
    RecreateWnd
  end
end;

{ TSplitter.xxxSplitting - These functions return the splitting state.
    If Splitting = True then the splitter is dragging. Property is
    protected, so it's useful for derived classes only. }
function TSplitter.GetSplitting: Boolean;
begin
  GetSplitting := (FSplitting <> 0)
end;

procedure TSplitter.SetSplitting(IsSplitting: Boolean);
begin
  FSplitting := Word(IsSplitting)
end;

{ TSplitter.xxxSplittingByMouse - Checks if splitting has been initiated by
    mouse. }
function TSplitter.GetSplittingByMouse: Boolean;
begin
  GetSplittingByMouse := (FSplitting and $8000 <> 0)
end;

procedure TSplitter.SetSplittingByMouse(SplittingByMouse: Boolean);
begin
  if SplittingByMouse then FSplitting := FSplitting or $8000
  else FSplitting := FSplitting and not $8000
end;

{ TSplitter.GetControlRect - Returns the bounding rectangle in screen
    coordinates }
function TSplitter.GetControlRect: TRect;
var
  rect: TRect;
begin
  rect := BoundsRect;
  MapWindowPoints(Parent.Handle, HWND_DESKTOP, rect, 2);
  GetControlRect := rect
end;

function TSplitter.GetControlOrigin: TPoint;
var
  rect: TRect;
begin
  rect := GetControlRect;
  GetControlOrigin := rect.TopLeft
end;

{ TSplitter.GetParentClientRect - Returns the parent's client
    rectangle in screen coordinates. }
function TSplitter.GetParentClientRect: TRect;
var
  rc: TRect;
begin
  rc := Parent.ClientRect;
  MapWindowPoints(Parent.Handle, HWND_DESKTOP, rc, 2);
  GetParentClientRect := rc
end;

function TSplitter.GetParentClientOrigin: TPoint;
begin
  GetParentClientOrigin := ClientOrigin
end;

procedure Register;
begin
  RegisterComponents('Samples', [TSplitter]);
end;

end.
